
/**
 * Generates a simple DISCOVER message
 */

#include "discover.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

/* PORT TO MAAS */
#define PORT  2535

static unsigned long MCAST_ADDR = 0;
static unsigned long scope_id = 0;
static struct sockaddr_in server;
static struct sockaddr_in client;
static int init_socket();
static int main_loop(void);
static int receive_msg (int socket);
static int send_discover(int socket);

enum{
    DISCOVER = 1,
    OFFER    = 2,
    REQUEST  = 3,
    RENEW    = 4,
    ACK      = 5,
    NAK      = 6,
    RELEASE  = 7,
    GETINFO  = 8
};

int discover(struct scope sc){
    MCAST_ADDR = (unsigned long)sc.addresses.end - 1;
    scope_id   = sc.addresses.start;
    main_loop();

    return 0;
}
/**
 * The main_loop function implements a infinite loop listening
 * for network data (should be OFFER msg) to arrive on the socket or for a
 * time out  event to occur.
 */ 
static int main_loop(){
    int sd = 0; /* The socket to send/receive mesasages to/from */
    int rc = 0; /* The number of descriptors contained in
                   the descriptor set */
    struct timeval tv;
    struct timeval tv_tmp;
    
    fd_set set;    
    
    sd = init_socket(); /* Init the socket */
    send_discover(sd);  /* Send DISCOVER msg */
    FD_ZERO(&set);      /* Empties the filedesriptor set */    
    FD_SET(sd,&set);    /* Add the socket */

     /* Set the timeout for listening */
    tv.tv_sec = 5;
    tv.tv_usec = 0;
    tv_tmp.tv_sec = tv.tv_sec;


    /* Wait for reply from MAAS's */
    for (;;){ 
        
        /* Select only among the read fd's(only one here) and with */         
        rc = select(FD_SETSIZE, &set, NULL, NULL, &tv);
                
        /* We have an error */
        if (rc < 0){
            fprintf(stderr,"select fails.\n");
            return -1;
        }

        /* We have a timeout */
        if(rc == 0){
            fprintf(stderr,"\nTimeout after %ld seconds.\n",tv_tmp.tv_sec);
            return 0;
        }else if (FD_ISSET(sd, &set)){
            /* Registered socket has received data */
            receive_msg(sd);
        }
    }
    return 0;
}

/**
 * Init the socket
 */
static int init_socket(){
    struct sockaddr_in tmp; 
    struct ip_mreq ipmq;
    char multicast_group[50];
    int sd              = 0;
    unsigned char loop  = 0;
    int ttl             = 16;

    /* Create socket */
    if ((sd = socket(AF_INET, SOCK_DGRAM, 0)) < 0){
        fprintf(stderr,"init_serversocket: socket failed (%s)\n",
                strerror(errno));
        exit(-6);
    }

    /* Fill in values in  sockaddr_in structure */
    bzero((char *)&server, sizeof(server));
    bzero((char *)&client, sizeof(client));
    server.sin_family      = AF_INET;
    server.sin_addr.s_addr = ntohl(MCAST_ADDR);
    server.sin_port        = htons(PORT);
    client.sin_family      = AF_INET;
    client.sin_addr.s_addr = INADDR_ANY;
    client.sin_port        = INADDR_ANY; 
    
    /* Bind address to socket */
    if (bind(sd, (struct sockaddr *)&client, sizeof(struct sockaddr)) < 0){
        fprintf(stderr,"can't bind local address");
        exit(-6);
    }
       
    /* Join multicast group */
    bzero((char *)&ipmq , sizeof(ipmq));
    bzero((char *)&tmp , sizeof(tmp));

    ipmq.imr_multiaddr.s_addr = ntohl(MCAST_ADDR);
    ipmq.imr_interface.s_addr = INADDR_ANY;
    
    if (setsockopt(sd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
                   &ipmq, sizeof(ipmq)) < 0){
        fprintf(stderr,
                "init_serversock: setsockopt IP_ADD_MEMBERSHIP failed (%s)\n",
                strerror(errno));
        exit(-6); 
    }

    /* Set TTL to traverse up to multiple routers. */
    if (setsockopt(sd, IPPROTO_IP,
                   IP_MULTICAST_TTL, &ttl, sizeof(ttl)) < 0){
        fprintf(stderr,
                "init_serversock: setsockopt IP_MULTICAST_TTL failed (%s)\n",
                strerror(errno));
        exit(-6); 
    }

    /* Disable loopback. */
    loop = 0;
    if (setsockopt(sd, IPPROTO_IP, IP_MULTICAST_LOOP,
                   &loop, sizeof(loop)) < 0){
        fprintf(stderr,
                "init_serversock: setsockopt IP_MULTICAST_LOOP failed (%s)\n",
                strerror(errno));
        exit(-6); 
    }
        
    
    tmp.sin_addr.s_addr = ipmq.imr_multiaddr.s_addr; 
    sprintf(multicast_group,"Joined multicast group %s:%d with TTL %d.%s",
            inet_ntoa(tmp.sin_addr), PORT, ttl, "\n");    
    printf("%s",multicast_group);
    return sd;
}


static int receive_msg (int sock){
    struct sockaddr_in maas;
    struct sockaddr_in from;
    int BUF_SIZE             = 1500;
    int len                  = 0;
   /*   int i                    = 0; */
    unsigned char uint8      = 0;
    unsigned short uint16    = 0;
    unsigned int uint32      = 0;
    time_t maas_current_time = 0;
    time_t start_time        = 0;
    time_t maas_offer_time   = 0;
    time_t valid             = 0;
   /*   int reportCount          = 0; */
/*      int requestCount         = 0; */
    char *buff               = (char *)malloc(sizeof(char)*1000);
    int fromlen              = sizeof(struct sockaddr_in);
    
    bzero((char *)&maas, sizeof(maas));
    bzero((char *)&from , sizeof(from));
    
    len = recvfrom(sock, buff, BUF_SIZE, 0,
                    (struct sockaddr *)&maas, &fromlen);

    if(len == -1) {
        fprintf(stderr,"receive_msg: Failed to read from socket\n");
        return -1;
    }

    /* Processing incomming message */   
    printf("\n");

    if(len < 12) {
        printf("Message length to short: len < 12 (%d)\n", len);
        return -1;
    }
    
    /* Get version */
    memcpy(&uint8, buff, 1);
    if (uint8 != 0){
        printf("obscure message received\n");
        return -1;
    }else{
        /*  printf("ver 0x%x, ", uint8); */
    }
    buff++;
    len--;

    /* Get message type */
    memcpy(&uint8, buff, 1);
    buff++;
    len--; 
    switch (uint8){   
    case DISCOVER:
        printf("Receiving DISCOVER message from %s, ignoring \n",
               inet_ntoa(maas.sin_addr));
        return 0;
        break;

    case OFFER:
        printf("Receiving OFFER message from %s.\n",
               inet_ntoa(maas.sin_addr));
        break;
        
    case REQUEST:
        printf("Receiving REQUEST message from %s, ignoring...\n",
               inet_ntoa(maas.sin_addr));
        return 0; 
        break;

    case RENEW:
        printf("Receiving RENEW message from %s, ignoring...\n",
               inet_ntoa(maas.sin_addr));
        return 0; 
        break;

    case ACK:
        printf("Receiving ACK message from %s, ignoring...\n",
               inet_ntoa(maas.sin_addr));
        return 0; 
        break;

    case NAK:
        printf("Receiving NAK message from %s, ignoring...\n",
               inet_ntoa(maas.sin_addr));
        return 0; 
        break;

    case RELEASE:
        printf("Receiving RELEASE message from %s, ignoring...\n",
               inet_ntoa(maas.sin_addr));
        return 0; 
        break;
        
    case GETINFO:
        printf("Receiving GETINFO message from %s, ignoring...\n",
               inet_ntoa(maas.sin_addr));
        return 0; 
        break;        
        
    default:
        printf("Receiving unknown message type ignoring...\n");
        return 0;

        break;
    }

    /* Get the rest of the header(of the DISCOVER msg) */
    /* addr_family */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
   /*   printf("addr fam 0x%x ", uint16); */
    buff = buff + 2;
    len = len - 2;

     /***************************************************/

    /* xid */
    memcpy(&uint32, buff, 4);
    uint32 = ntohl(uint32);
    printf("Transaction id:\t%d\n",uint32);
    buff = buff + 4;
    len = len  - 4;

     /***************************************************/

    /* lease time(code 1) */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
    if (uint16 != 1){
        printf("Invalid option. Lease time code %d\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;

    /* len should be 4 */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
     if (uint16 != 4){
        printf("Invalid option. Lease len %d.\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;
    
    /* Get Lease time */
    memcpy(&uint32, buff, 4);
    maas_offer_time = ntohl(uint32);
    buff = buff + 4;
    len = len - 4;
    valid = time(NULL) + maas_offer_time;
    printf("Lease time: \t%s",ctime(&valid));

    /***************************************************/

    /* server id(code 2) */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
    if (uint16 != 2){
        printf("Invalid option. Server id code %d\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;

    /* len should be 6 in this version */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
     if (uint16 != 6){
        printf("Invalid option. Serverid len %d\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;

    /* address family */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
     if (uint16 != 1){
        printf("Invalid option. Not supported IP version %d.\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;    

    /* the server id */
    memcpy(&uint32, buff, 4);
    from.sin_addr.s_addr = uint32;   
    buff = buff + 4;
    len = len - 4;
    printf("Server id: \t%s\n", inet_ntoa(from.sin_addr));

    /***************************************************/


    /* lease id (code 3) */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
    if (uint16 != 3){
        printf("Invalid option. Lease id code %d\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;

    /* we sent a lease id with len 1 */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
    if (uint16 != 1){
        printf("Invalid option. Lease id len %d\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;

    memcpy(&uint8, buff, 1);
    printf("Lease id: \t%d\n",uint8);
    buff++;
    len--;
    

    /***************************************************/

    /* multicast scope(code 4) */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
    if (uint16 != 4){
        printf("Invalid option. Multicast scope code %d.\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;
    
    /* len should at least 1 */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
     if (uint16 < 1){
        printf("Invalid option. Multicast scope len %d\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;

    /* the scope OFFER */
    memcpy(&uint32, buff, 4);
    from.sin_addr.s_addr = uint32;   
    buff = buff + 4;
    len = len - 4;
    printf("Scope: \t\t%s\n", inet_ntoa(from.sin_addr));

    /***************************************************/
    
    /* start time(code 6) */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
    if (uint16 != 6){
        printf("Invalid option. Start time code %d\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;

    /* len should be 4  */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
     if (uint16 !=4){
        printf("Invalid option. Start time len %d\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;

    /* the time */
    memcpy(&uint32, buff, 4);
    start_time = ntohl(uint32);
    buff = buff + 4;
    len = len - 4;
    printf("Start time: \t%s",ctime(&start_time));

    /***************************************************/
    
    /* list of address ranges */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
    if (uint16 != 10){
        printf("Invalid option. Address range code %d\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;

    /* len */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
    if (uint16 <= 0){
        printf("Invalid option. Address range len %d\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;

    /* Skipp address ranges */      
    buff = buff + uint16;
    len = len - uint16;

    /***************************************************/

    /* current time */    
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
    if (uint16 != 11){
        printf("Invalid option. Current time code %d\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;

    /* len */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
    if (uint16 != 4){
        printf("Invalid option. Current time len %d\n",uint16);
        return 0;
    }
    buff = buff + 2;
    len = len - 2;

    /* the time */
    memcpy(&uint32, buff, 4);
    maas_current_time = ntohl(uint32);
    buff = buff + 4;
    len = len - 4;
    printf("Current time: \t%s",ctime(&maas_current_time));

    /***************************************************/

    /* end */
    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
    if (uint16 != 0){
        printf("Invalid option. Expecting end code 0\n");
        return 0;
    }
    buff = buff + 2;
    len = len - 2;

    memcpy(&uint16, buff, 2);
    uint16 = ntohs(uint16);
    if (uint16 != 0){
        printf("Invalid option. Expecting end len 0\n");
        return 0;
    }
    buff = buff + 2;
    len = len - 2;
    return 0;
}


/* Send discover message */
static int send_discover(int socket){
    int size              = 0;
    unsigned char uint8   = 0;
    unsigned short uint16 = 0;
    unsigned int uint32   = 0;    
    struct sockaddr_in tmp; 

    char *data = (char *)malloc(sizeof(char)*1000);
    char *p = data;
    
    printf("Constructing DISCOVER message: ");
    /* Format DISCOVER */
    /* version */
    uint8 = 0;
    printf("ver 0x%x, ", uint8);
    memcpy(data, &uint8, 1);
    data++;
    size = size + 1;
    
    /* type */
    uint8 = DISCOVER;
    memcpy(data, &uint8, 1);
    data++;
    size = size + 1;
    
    /* addrfamily */
    printf("addr fam 0x%x\n", 1);
    uint16 = htons(1);    
    memcpy(data, &uint16, 2);
    data = data + 2;
    size = size + 2;

    /* xid */
    uint32 = 133;
    printf("Transaction id:\t%d\n",uint32);
    uint32 = htonl(uint32);
    memcpy(data,  &uint32, 4);
    data = data + 4;
    size = size + 4;

    /* Fromat options */
    /* Lease id */
    uint16 = 3;   
    uint16 = htons(uint16);
    memcpy(data, &uint16, 2);
    data = data + 2;
    size = size + 2;
    
    uint16 = htons(1);      /* len = 1 */
    memcpy(data, &uint16, 2);
    data = data + 2;
    size = size + 2;
    
    uint8 = 13;             /* 'unique' lease id */
    printf("Lease id: \t%d\n",uint8);
    memcpy(data, &uint8, 1);
    data++;
    size++;

    /* Multicast scope */
    uint16 = htons(4);
    memcpy(data, &uint16, 2);
    data = data + 2;
    size = size + 2;
    uint16 = htons(4);        /* len = 4*/
    memcpy(data, &uint16, 2);
    data = data + 2;
    size = size + 2;    

    tmp.sin_addr.s_addr = ntohl(scope_id);
    printf("Scope id: \t%s\n", inet_ntoa(tmp.sin_addr));
    uint32 =  htonl(scope_id); /* scope id */
    memcpy(data, &uint32, 4);
    data = data + 4;
    size = size + 4;

    /* end */
    uint16 = htons(0);
    memcpy(data, &uint16, 2);
    data = data + 2;
    size = size + 2;
    uint16 = htons(0); /* len = 0 */
    memcpy(data, &uint16, 2);
    data = data + 2;
    size = size + 2;

    tmp.sin_addr.s_addr = ntohl(MCAST_ADDR);    
    printf("Sending the DISCOVER message. (%d bytes) to %s\n",
           size,inet_ntoa(tmp.sin_addr));
    /* Send DISCOVER message to multicast group */
    if(sendto(socket, p, size, 0, (struct sockaddr *)&server, sizeof(server)) < size) {
       fprintf(stderr,"send_DISCOVER: sendto failed (%s)\n",
               strerror(errno));
       printf("size: %d\n",size);
    }    
    return 0;
    return 0;
}

